// Climate Framework for Uncertainty, Negotiation and Distribution
// Version 3.6
// (c) David Anthoff and Richard S.J. Tol
// http://www.fund-model.org
using System;
using System.Collections.Generic;
using System.Text;

namespace Esmf
{
    public class ModelFields : IDisposable
    {
        private bool _disposed = false;

        public class Field
        {
            public string ComponentName { get; set; }
            public string FieldName { get; set; }
            public object Values { get; set; }
        }

        class FieldValue<T>
        {
            private T _value;
            public T Value
            {
                get { return _value; }
                set { _value = value; }
            }
        }

        private class DelegateToParameter1Dimensional<D1, T> : IParameter1Dimensional<D1, T>
        {
            Func<D1, T> _delegate;

            public DelegateToParameter1Dimensional(Func<D1, T> function)
            {
                _delegate = function;
            }

            T IParameter1Dimensional<D1, T>.this[D1 index]
            {
                get { return _delegate(index); }
            }

            public IEnumerable<Parameter1DimensionalMember<D1, T>> GetEnumerator()
            {
                throw new NotImplementedException();
            }
        }


        private class DelegateToParameter2Dimensional<D1,D2,T> : IParameter2Dimensional<D1,D2,T>
        {
            private ModelFields _parent;
            Func<D1, D2, T> _delegate;
            HashSet<D1> _d1Members = new HashSet<D1>();
            HashSet<D2> _d2Members = new HashSet<D2>();

            public DelegateToParameter2Dimensional(Func<D1, D2, T> function, ModelFields parent)
            {
                _delegate = function;
                _parent = parent;
            }

            T IParameter2Dimensional<D1, D2, T>.this[D1 index1, D2 index2]
            {
                get 
                {
                    _d1Members.Add(index1);
                    _d2Members.Add(index2);

                    return _delegate(index1, index2); 
                }
            }

            #region IParameter2Dimensional<D1,D2,T> Members


            public IEnumerable<Parameter2DimensionalMember<D1, D2, T>> GetEnumerator()
            {
                foreach (var d1 in _d1Members)
                {
                    foreach (var d2 in _d2Members)
                    {
                        yield return new Parameter2DimensionalMember<D1, D2, T>(d1, d2, _delegate(d1, d2), _parent.AttachedDimensions);
                    }
                }
            }

            #endregion
        }

        List<Tuple<string, string>> _variablesKeysAlreadyUsed = new List<Tuple<string, string>>();

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (!_disposed)
            {
                if (disposing)
                {
                    foreach (var p in _dimensionalVariables.Values)
                    {
                        if (p is IDisposable)
                        {
                            ((IDisposable)p).Dispose();
                        }
                    }
                }

                _attachedDimensions = null;
                _dimensionalVariables = null;
                _nonDimensionalVariables = null;
                _nonDimensionalVariablesGetterDelegates = null;
                _nonDimensionalVariablesSetterDelegates = null;
                _variablesKeysAlreadyUsed = null;

                _disposed = true;
            }
        }

        ~ModelFields()
        {
            Dispose(false);
        }

        #region NonDimensionalVariables
        Dictionary<Type, object> _nonDimensionalVariables = new Dictionary<Type, object>();
        Dictionary<Tuple<string, string>, object> _nonDimensionalVariablesGetterDelegates = new Dictionary<Tuple<string, string>, object>();
        Dictionary<Tuple<string, string>, object> _nonDimensionalVariablesSetterDelegates = new Dictionary<Tuple<string, string>, object>();

        public void AddNonDimensionalVariable<T>(string componentName, string fieldName) where T : struct
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            var key = new Tuple<string,string>(componentName, fieldName);
            Dictionary<Tuple<string, string>, FieldValue<T?>> d;

            if (_nonDimensionalVariables.ContainsKey(typeof(T)))
            {
                d = (Dictionary<Tuple<string, string>, FieldValue<T?>>)_nonDimensionalVariables[typeof(T)];
            }
            else
            {
                d = new Dictionary<Tuple<string, string>, FieldValue<T?>>();
                _nonDimensionalVariables.Add(typeof(T), d);
            }

            if (_variablesKeysAlreadyUsed.Contains(key))
            {
                throw new ArgumentException();
            }

            d.Add(key, new FieldValue<T?>());

            _variablesKeysAlreadyUsed.Add(key);

            FieldValue<T?> v = d[key];

            _nonDimensionalVariablesGetterDelegates.Add(key, (NonDimensionalFieldGetter<T>)delegate { return (T)v.Value; });
            _nonDimensionalVariablesSetterDelegates.Add(key, (NonDimensionalFieldSetter<T>)delegate(T value) { v.Value = value; });
        }

        public void AddNonDimensionalVariable<T>(string componentName, string fieldName, T value) where T : struct
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            AddNonDimensionalVariable<T>(componentName, fieldName);

            Dictionary<Tuple<string, string>, FieldValue<T?>> d = (Dictionary<Tuple<string, string>, FieldValue<T?>>)_nonDimensionalVariables[typeof(T)];

            d[new Tuple<string, string>(componentName, fieldName)].Value = value;
        }

        public void LoadNonDimensionalVariableFromParameters<T>(string componentName, string fieldName, ParameterValues parameters) where T : struct
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            T value = ((ParameterValueNonDimensional<T>)parameters[fieldName]).Value;

            AddNonDimensionalVariable<T>(componentName, fieldName, value);
        }

        public void LoadOneDimensionalVariableFromParameters<D1, T>(string componentName, string fieldName, ParameterValues parameters) where T : struct
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            var p = (ParameterValue1Dimensional<T>) parameters[fieldName];

            if (typeof(D1) == typeof(Timestep))
            {
                var value = new VariableWithTimeDimension<T>(this, p);

                Add1DimensionalParameter<Timestep, T>(componentName, fieldName, value);
            }
            else if (typeof(D1).BaseType == typeof(Enum))
            {
                var value = new Variable1Dimensional<D1, T>(this, p);

                Add1DimensionalParameter<D1, T>(componentName, fieldName, value);
            }
            else
                throw new ArgumentException("Unknown dimension type");            
        }

        public void LoadTwoDimensionalVariableFromParameters<D1, D2, T>(string componentName, string fieldName, ParameterValues parameters, string parameterName) where T : struct
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            var p = (ParameterValue2Dimensional<T>)parameters[parameterName];

            if (typeof(D1) == typeof(Timestep) && typeof(D2).BaseType==typeof(System.Enum))
            {
                var value = new Variable2DimensionalTime<D2, T>(this, p);

                Add2DimensionalParameter<Timestep, D2, T>(componentName, fieldName, value);
            }
            else if (typeof(D1).BaseType == typeof(Enum) && typeof(D2).BaseType==typeof(Enum))
            {
                var value = new Variable2Dimensional<D1,D2, T>(this, p);

                Add2DimensionalParameter<D1, D2, T>(componentName, fieldName, value);
            }
            else
                throw new ArgumentException("Unknown dimension type");
        }

        public void LoadTwoDimensionalVariableFromParameters<D1, D2, T>(string componentName, string fieldName, ParameterValues parameter) where T : struct
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            LoadTwoDimensionalVariableFromParameters<D1, D2, T>(componentName, fieldName, parameter, fieldName);
        }

        public object GetNonDimensionalVariableGetter(string componentName, string fieldName)
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            return _nonDimensionalVariablesGetterDelegates[new Tuple<string, string>(componentName, fieldName)];
        }

        public NonDimensionalFieldGetter<T> GetNonDimensionalVariableGetter<T>(string componentName, string fieldname)
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            return (NonDimensionalFieldGetter<T>)GetNonDimensionalVariableGetter(componentName, fieldname);
        }

        public object GetNonDimensionalVariableSetter(string componentName, string fieldName)
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            return _nonDimensionalVariablesSetterDelegates[new Tuple<string, string>(componentName, fieldName)];
        }

        public NonDimensionalFieldSetter<T> GetNonDimensionalVariableSetter<T>(string componentName, string fieldName)
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            return (NonDimensionalFieldSetter<T>)GetNonDimensionalVariableSetter(componentName, fieldName);
        }

        public void SetNonDimensionalFieldValue<T>(string componentName, string fieldName, T value)
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            NonDimensionalFieldSetter<T> setter = GetNonDimensionalVariableSetter<T>(componentName, fieldName);
            setter(value);
        }

        #endregion

        #region DimensionalVariables

        Dictionary<Tuple<string, string>, object> _dimensionalVariables = new Dictionary<Tuple<string, string>, object>();

        public void Add1DimensionalVariable<D1, T>(string componentName, string fieldName)
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            var key = new Tuple<string, string>(componentName, fieldName);

            if (_variablesKeysAlreadyUsed.Contains(key))
            {
                throw new ArgumentException();
            }

            object v;
            if (typeof(D1) == typeof(Timestep))
            {
                v = new VariableWithTimeDimension<T>(this);
            }
            else
            {
                v = new Variable1Dimensional<D1, T>(this);
            }

            _dimensionalVariables.Add(key, v);
            _variablesKeysAlreadyUsed.Add(key);
        }


        public void Add2DimensionalVariable<D1, D2, T>(string componentName, string fieldName)
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            var key = new Tuple<string, string>(componentName, fieldName);

            if (_variablesKeysAlreadyUsed.Contains(key))
            {
                throw new ArgumentException();
            }

            object v;
            if (typeof(D1) == typeof(Timestep))
            {
                v = new Variable2DimensionalTime<D2, T>(this);
            }
            else
            {
                v = new Variable2Dimensional<D1, D2, T>(this);
            }

            _dimensionalVariables.Add(key, v);
            _variablesKeysAlreadyUsed.Add(key);
        }

        public void Add1DimensionalParameter<D1, T>(string componentName, string fieldName, IParameter1Dimensional<D1, T> parameter)
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            var key = new Tuple<string, string>(componentName, fieldName);

            if (_variablesKeysAlreadyUsed.Contains(key))
            {
                throw new ArgumentException();
            }

            _dimensionalVariables.Add(key, parameter);
            _variablesKeysAlreadyUsed.Add(key);
        }

        public void Add1DimensionalParameter<D1,T>(string componentName, string fieldName, Func<D1,T> parameter)
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            Add1DimensionalParameter(componentName, fieldName, new DelegateToParameter1Dimensional<D1,T>(parameter));
        }

        public void Add2DimensionalParameter<D1, D2, T>(string componentName, string fieldName, IParameter2Dimensional<D1, D2, T> parameter)
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            var key = new Tuple<string, string>(componentName, fieldName);

            if (_variablesKeysAlreadyUsed.Contains(key))
            {
                throw new ArgumentException();
            }

            _dimensionalVariables.Add(key, parameter);
            _variablesKeysAlreadyUsed.Add(key);
        }

        public void Add2DimensionalParameter<D1, D2, T>(string componentName, string fieldName, Func<D1,D2,T> parameter)
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            Add2DimensionalParameter(componentName, fieldName, new DelegateToParameter2Dimensional<D1, D2, T>(parameter, this));
        }

        public void ConnectParameterToVariable(string parameterComponentName, string parameterFieldName, string variableComponentName, string variableFieldName)
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            var key = new Tuple<string, string>(parameterComponentName, parameterFieldName);

            if (_variablesKeysAlreadyUsed.Contains(key))
            {
                throw new ArgumentException();
            }

            if (_dimensionalVariables.ContainsKey(new Tuple<string, string>(variableComponentName, variableFieldName)))
            {
                _dimensionalVariables.Add(key, _dimensionalVariables[new Tuple<string, string>(variableComponentName, variableFieldName)]);
            }
            else
            {
                throw new InvalidOperationException();
            }
            _variablesKeysAlreadyUsed.Add(key);
        }

        public void ConnectParameterToVariableNonDimensional<T>(string parameterComponentName, string parameterFieldName, string variableComponentName, string variableFieldName) where T : struct
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            var key = new Tuple<string, string>(parameterComponentName, parameterFieldName);
            var sourceKey = new Tuple<string, string>(variableComponentName, variableFieldName);

            if (_variablesKeysAlreadyUsed.Contains(key))
            {
                throw new ArgumentException();
            }

            Dictionary<Tuple<string, string>, FieldValue<T?>> d = (Dictionary<Tuple<string, string>, FieldValue<T?>>)_nonDimensionalVariables[typeof(T)];

            _variablesKeysAlreadyUsed.Add(key);

            FieldValue<T?> v = d[sourceKey];

            _nonDimensionalVariablesGetterDelegates.Add(key, (NonDimensionalFieldGetter<T>)delegate { return (T)v.Value; });
        }


        public object GetDimensionalField(string componentName, string fieldName)
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            var key = new Tuple<string, string>(componentName, fieldName);
            if(!_dimensionalVariables.ContainsKey(key))
            {
                throw new ArgumentOutOfRangeException(String.Format("{0}.{1} is not in the list of dimensional fields", componentName, fieldName));
            }

            return _dimensionalVariables[key];
        }

        public void Set1DimensionalParameter<D1, T>(string componentName, string fieldName, Func<D1, T> parameter)
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            var fieldKey = new Tuple<string, string>(componentName, fieldName);
            if (!_dimensionalVariables.ContainsKey(fieldKey))
            {
                throw new InvalidOperationException();
            }

            _dimensionalVariables[fieldKey] = new DelegateToParameter1Dimensional<D1, T>(parameter);
        }

        public void Set1DimensionalParameter<D1, T>(string componentName, string fieldName, ParameterValues parameters, string parameterName) where T : struct
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            var fieldKey = new Tuple<string, string>(componentName, fieldName);
            if (!_dimensionalVariables.ContainsKey(fieldKey))
            {
                throw new InvalidOperationException();
            }

            var p = (ParameterValue1Dimensional<T>)parameters[parameterName];

            if (typeof(D1) == typeof(Timestep))
            {
                var value = new VariableWithTimeDimension<T>(this, p);

                _dimensionalVariables[fieldKey] = value;
            }
            else if (typeof(D1).BaseType == typeof(Enum))
            {
                var value = new Variable1Dimensional<D1, T>(this, p);

                _dimensionalVariables[fieldKey] = value;
            }
            else
                throw new ArgumentException("Unknown dimension type");
        }

        public void Set1DimensionalParameter<D1, T>(string componentName, string fieldName, ParameterValues parameters) where T : struct
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            Set1DimensionalParameter<D1, T>(componentName, fieldName, parameters, fieldName);
        }


        public void Set2DimensionalParameter<D1, D2, T>(string componentName, string fieldName, Func<D1, D2, T> parameter)
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            var fieldKey = new Tuple<string, string>(componentName, fieldName);
            if (!_dimensionalVariables.ContainsKey(fieldKey))
            {
                throw new InvalidOperationException();
            }

            _dimensionalVariables[fieldKey] = new DelegateToParameter2Dimensional<D1, D2, T>(parameter, this);
        }

        public void Set2DimensionalParameter<D1, D2, T>(string componentName, string fieldName, IParameter2Dimensional<D1, D2, T> parameter)
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            var fieldKey = new Tuple<string, string>(componentName, fieldName);
            if (!_dimensionalVariables.ContainsKey(fieldKey))
            {
                throw new InvalidOperationException();
            }

            _dimensionalVariables[fieldKey] = parameter;
        }

        public void Set2DimensionalParameter<D1, D2, T>(string componentName, string fieldName, ParameterValues parameters, string parameterName) where T : struct
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            var fieldKey = new Tuple<string, string>(componentName, fieldName);

            if (!_dimensionalVariables.ContainsKey(fieldKey))
            {
                throw new InvalidOperationException();
            }

            var p = (ParameterValue2Dimensional<T>)parameters[parameterName];

            if (typeof(D1) == typeof(Timestep) && typeof(D2).BaseType == typeof(System.Enum))
            {
                var value = new Variable2DimensionalTime<D2, T>(this, p);

                _dimensionalVariables[fieldKey] = value;
            }
            else if (typeof(D1).BaseType == typeof(Enum) && typeof(D2).BaseType == typeof(Enum))
            {
                var value = new Variable2Dimensional<D1, D2, T>(this, p);

                _dimensionalVariables[fieldKey] = value;
            }
            else
                throw new ArgumentException("Unknown dimension type");
        }

        public void Set2DimensionalParameter<D1, D2, T>(string componentName, string fieldName, ParameterValues parameter) where T : struct
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            Set2DimensionalParameter<D1, D2, T>(componentName, fieldName, parameter, fieldName);
        }


        #endregion

        #region Attached dimensions

        private List<string> _attachedDimensions = new List<string>();

        public IList<string> AttachedDimensions
        {
            get
            {
                if (_disposed) throw new ObjectDisposedException("ModelFields");

                return _attachedDimensions;
            }
        }

        #endregion

        public IEnumerable<Field> GetDimensionalFieldsOperator()
        {
            if (_disposed) throw new ObjectDisposedException("ModelFields");

            foreach (var f in _dimensionalVariables.Keys)
            {
                yield return new Field() { ComponentName = f.Item1, FieldName = f.Item2, Values = _dimensionalVariables[f] };
            }
        }
    }
}
