﻿// Climate Framework for Uncertainty, Negotiation and Distribution
// Version 3.6
// (c) David Anthoff and Richard S.J. Tol
// http://www.fund-model.org
using System.Collections.Generic;
using System.Configuration;
using System.IO;
using System;
using Esmf;
using System.Threading.Tasks;
using System.Threading;
using System.Globalization;
using System.Collections.Concurrent;
using System.Linq;
using System.Linq.Expressions;


namespace Fund
{

    public enum MarginalGas { C, CH4, N2O, SF6 }

    public class TMarginalRun
    {
        private TextWriter m_YearRegionSectorWeightingSchemeCsv;
        private TextWriter m_AggregateDamageCsv;
        private TextWriter m_GlobalInputCsv;
        private TextWriter m_RegionInputCsv;
        private TextWriter m_YearInputCsv;
        private TextWriter m_RegionYearInputCsv;

        private Run _run;
        private MarginalGas _gas;
        private Timestep _emissionyear;
        private string _outputPath;
        private Parameters _parameters;

        private Random _rand;

        public TextWriter AggregateDamageCsv { get { return m_AggregateDamageCsv; } set { m_AggregateDamageCsv = value; } }
        public TextWriter YearRegionSectorWeightingSchemeCsv { get { return m_YearRegionSectorWeightingSchemeCsv; } set { m_YearRegionSectorWeightingSchemeCsv = value; } }
        public TextWriter GlobalInputCsv { get { return m_GlobalInputCsv; } set { m_GlobalInputCsv = value; } }
        public TextWriter RegionInputCsv { get { return m_RegionInputCsv; } set { m_RegionInputCsv = value; } }
        public TextWriter YearInputCsv { get { return m_YearInputCsv; } set { m_YearInputCsv = value; } }
        public TextWriter RegionYearInputCsv { get { return m_RegionYearInputCsv; } set { m_RegionYearInputCsv = value; } }
        public Action<FundWorkflow> AdditionalInitCode { get; set; }

        public TMarginalRun(Run runData, MarginalGas gas, Timestep emissionyear, string outputPath, Parameters parameters, Random rand = null)
        {
            _parameters = parameters;
            _run = runData;
            _gas = gas;
            _emissionyear = emissionyear;
            _outputPath = outputPath;
            _rand = rand;
        }

        // DA: Run the model once, write away income (GDP), population, market impacts, non market
        // impacts (prn files)
        // Run the model again, with a slightly higher emissions, additional emissions are in the file
        // addco2.001 (million ton of carbons) Write the results of the second run out to (out files)
        public double[] Run()
        {
            WeightingCombination[] i_weightingCombinations;

            Fund28LegacyWeightingCombinations.GetWeightingCombinationsFromName(_run.WeightingCombination, out i_weightingCombinations, _emissionyear);

            var sccResults = new ConcurrentBag<double>[i_weightingCombinations.Length];
            sccResults = new ConcurrentBag<double>[i_weightingCombinations.Length];

            for (int l = 0; l < i_weightingCombinations.Length; l++)
                sccResults[l] = new ConcurrentBag<double>();

            Console.Write("Best guess run ");

            var parameters = _parameters.GetBestGuess();

            var res = DoOneRun(0, i_weightingCombinations, parameters);

            int currentRun = 0;

            int batchSize = 100;

            int batchCount = (_run.MonteCarloRuns / batchSize) + 1;

            for (int currentBatch = 0; currentBatch < batchCount; currentBatch++)
            {
                int runsForThisBatch = currentBatch == batchCount - 1 ? _run.MonteCarloRuns % batchSize : batchSize;

                Parallel.ForEach(
                    _parameters.GetRandom(_rand, runsForThisBatch, currentBatch * batchSize),
                    () =>
                    {
                        Thread.CurrentThread.CurrentCulture = CultureInfo.InvariantCulture;
                        Thread.CurrentThread.Priority = ThreadPriority.BelowNormal;
                        return 0;
                    },
                    (p, pls, dummy) =>
                    {
                        int tempCurrentCount = Interlocked.Increment(ref currentRun);
                        Console.Write("\rRun {0}                ", tempCurrentCount);

                        res = DoOneRun(p.RunId.Value, i_weightingCombinations, p);

                        p.Dispose();

                        if (!double.IsNaN(res[0]) && !double.IsInfinity(res[0]))
                        {
                            for (int l = 0; l < i_weightingCombinations.Length; l++)
                                sccResults[l].Add(res[l]);
                        }

                        return 0;
                    },
                    (dummy) => { });
                GC.Collect();
            }
            Console.WriteLine();

            double[] res2 = new double[i_weightingCombinations.Length];

            if (_run.MonteCarloRuns > 0)
            {
                for (int l = 0; l < i_weightingCombinations.Length; l++)
                {
                    res2[l] = sccResults[l].Average();

                    WriteAggregateDamage(-1, l, res2[l], i_weightingCombinations);
                }
            }

            return res2;
        }

        public double[] DoOneRun(int RunId, WeightingCombination[] i_weightingCombinations, ParameterValues parameters)
        {
            ModelOutput i_output2;
            Damages i_marginalDamages;
            double i_aggregatedDamage;
            ModelOutput i_output1;

            // Create Output object for run 1, set addmp to 0 so that
            // the extra greenhouse gases are not emitted and then run
            // the model
            i_output1 = new ModelOutput();

            FundWorkflow f1 = new FundWorkflow(parameters);

            if (AdditionalInitCode != null)
                AdditionalInitCode(f1);

            f1.ModelFields.SetNonDimensionalFieldValue("climate", "addmp", false);

            f1.Run(i_output1);
            f1.ModelFields.Dispose();

            // Create Output object for run 2, set addmp to 1 so that
            // the extra greenhouse gases for the marginal run are
            // emitted and then run the model
            i_output2 = new ModelOutput();

            FundWorkflow f2 = new FundWorkflow(parameters);

            if (AdditionalInitCode != null)
                AdditionalInitCode(f2);

            f2.ModelFields.SetNonDimensionalFieldValue("fundold2climate", "emissionPeriod", _emissionyear);
            f2.ModelFields.SetNonDimensionalFieldValue("fundold2climate", "marginalGas", _gas);
            f2.ModelFields.SetNonDimensionalFieldValue("climate", "addmp", true);

            f2.Run(i_output2);
            f2.ModelFields.Dispose();

            Fund28LegacyWeightingCombinations.GetWeightingCombinationsFromName(_run.WeightingCombination, out i_weightingCombinations, _emissionyear);

            // Take out growth effect effect of run 2 by transforming
            // the damage from run 2 into % of GDP of run 2, and then
            // multiplying that with GDP of run 1
            for (int year = 1; year < LegacyConstants.NYear; year++)
            {
                for (int region = 1; region <= LegacyConstants.NoReg; region++)
                {
                    for (int sector = 0; sector < LegacyConstants.NoSector; sector++)
                    {
                        i_output2.Damages[year, region, (Sector)sector] = (i_output2.Damages[year, region, (Sector)sector] / i_output2.Incomes[year,region]) * i_output1.Incomes[year, region];
                    }
                }
            }

            // Calculate the marginal damage between run 1 and 2 for each
            // year/region/sector
            i_marginalDamages = Damages.CalculateMarginalDamage(i_output1.Damages, i_output2.Damages);

            double[] i_weightedAggregatedDamages = new double[i_weightingCombinations.Length];

            for (int i = 0; i < i_weightingCombinations.Length; i++)
            {
                i_weightingCombinations[i].CalculateWeights(i_output1);
                i_aggregatedDamage = i_weightingCombinations[i].AddDamagesUp(i_marginalDamages, _run.YearsToAggregate, _emissionyear);

                i_weightedAggregatedDamages[i] = i_aggregatedDamage;

                WriteAggregateDamage(RunId, i, i_aggregatedDamage, i_weightingCombinations);

                // Console.Write(i_weightingCombinations[i].Name + ": ");
                // Console.WriteLine(Convert.ToString(i_aggregatedDamage));
            }

            if (m_YearRegionSectorWeightingSchemeCsv != null)
            {
                foreach(var i_Damage in i_marginalDamages)
                {
                    if ((i_Damage.Year >= _emissionyear.Value) && (i_Damage.Year < _emissionyear.Value + _run.YearsToAggregate))
                    {
                        for (int k = 0; k < i_weightingCombinations.Length; k++)
                            WriteMarginalDamage(RunId, i_Damage, k, i_weightingCombinations[k][i_Damage.Year, i_Damage.Region], i_weightingCombinations);
                    }
                }
            }

            return i_weightedAggregatedDamages;
        }

        public void WriteMarginalDamage(int RunId, Damage i_Damage, int WeightingschemeId, double Weight, WeightingCombination[] WeightingCombinations)
        {

            if (m_YearRegionSectorWeightingSchemeCsv != null)
            {
                int gas = Convert.ToInt32(_gas);

                m_YearRegionSectorWeightingSchemeCsv.WriteLine(
                    (_run.OutputVerbal ? _run.Scenario.Name : _run.Scenario.Id.ToString()) +
                    ";" +
                    (_run.OutputVerbal ? (RunId == 0 ? "Best guess" : RunId == -1 ? "Mean" : RunId.ToString()) : RunId.ToString()) +
                    ";" +
                    (_run.OutputVerbal ? (gas == 0 ? "C" : gas == 1 ? "CH4" : gas == 2 ? "N2O" : gas == 3 ? "SF6" : "ERROR") : ((int)_gas).ToString()) +
                    ";" +
                    _emissionyear.ToString() +
                    ";" +
                    (i_Damage.Year + 1950).ToString() +
                    ";" +
                    (_run.OutputVerbal ? LegacyRegionNames.GetName(i_Damage.Region) : i_Damage.Region.ToString()) +
                    ";" +
                    (_run.OutputVerbal ? Enum.GetName(typeof(Sector), i_Damage.Sector) : ((int)i_Damage.Sector).ToString()) +
                    ";" +
                    (_run.OutputVerbal ? WeightingCombinations[WeightingschemeId].Name : WeightingschemeId.ToString()) +
                    ";" +
                    (i_Damage.DamageValue * Weight).ToString("f15")
                    );
            }
        }

        public void WriteAggregateDamage(int RunId, int WeightingschemeId, double Damage, WeightingCombination[] WeightingCombinations)
        {           
            if (!_run.CalculateMeanForMonteCarlo && (RunId == -1))
                return;
            else if (!_run.OutputAllMonteCarloRuns && (RunId > 0))
                return;

            if (m_AggregateDamageCsv != null)
            {
                int gas = (int)_gas;

                m_AggregateDamageCsv.WriteLine(
                    (_run.OutputVerbal ? _run.Scenario.Name : _run.Scenario.Id.ToString()) +
                    ";" +
                    (_run.OutputVerbal ? (gas == 0 ? "C" : gas == 1 ? "CH4" : gas == 2 ? "N2O" : gas == 3 ? "SF6" : "ERROR") : ((int)_gas).ToString()) +
                    ";" +
                    _emissionyear.ToString() +
                    ";" +
                    (_run.OutputVerbal ? (RunId == 0 ? "Best guess" : RunId == -1 ? "Mean" : RunId.ToString()) : RunId.ToString()) +
                    ";" +
                    (_run.OutputVerbal ? WeightingCombinations[WeightingschemeId].Name : WeightingschemeId.ToString()) +
                    ";" +
                    Damage.ToString("f15")
                    );


            }
        }
    }
}