﻿// Climate Framework for Uncertainty, Negotiation and Distribution
// Version 3.6
// (c) David Anthoff and Richard S.J. Tol
// http://www.fund-model.org
using System.Collections;
using System.Collections.Generic;

namespace Fund
{

    // Class that holds all output from one model run
    public class ModelOutput
    {
        private Damages m_damages = new Fund.Damages();
        private Incomes m_incomes = new Fund.Incomes();
        private Populations m_populations = new Fund.Populations();


        public Damages Damages { get { return m_damages; } set { m_damages = value; } }
        public Incomes Incomes { get { return m_incomes; } set { m_incomes = value; } }
        public Populations Populations { get { return m_populations; } set { m_populations = value; } }
    }

    public class Damage
    {
        public int Year { get; set; }
        public int Region { get; set; }
        public int Sector { get; set; }
        public double DamageValue { get; set; }
    }

    // Collection of all Damages for one model run
    public class Damages : IEnumerable<Damage>
    {
        private Esmf.JaggedArrayWrapper3<double> _damages = new Esmf.JaggedArrayWrapper3<double>(LegacyConstants.NYear, LegacyConstants.NoReg, LegacyConstants.NoSector);

        public double this[int year,int region, Sector sector]
        {
            get
            {
                return _damages[year - 1, region - 1, (int)sector]; 
            }
            set
            {
                _damages[year-1, region-1, (int)sector] = value;
            }
        }

        public static Damages CalculateMarginalDamage(Damages Damages1, Damages Damages2)
        {
            Damages i_marginalDamages= new Damages();

            for (int year = 0; year < LegacyConstants.NYear; year++)
            {
                for (int region = 0; region < LegacyConstants.NoReg; region++)
                {
                    for (int sector = 0; sector < LegacyConstants.NoSector; sector++)
                    {
                        i_marginalDamages._damages[year, region, sector] = (Damages2._damages[year, region, sector] - Damages1._damages[year, region, sector]) / 10000000;
                    }
                }
            }

            return i_marginalDamages;
        }

        public Damages()
        {
            
        }


        public IEnumerator<Damage> GetEnumerator()
        {
            for (int year = 0; year < LegacyConstants.NYear; year++)
            {
                for (int region = 0; region < LegacyConstants.NoReg; region++)
                {
                    for (int sector = 0; sector < LegacyConstants.NoSector; sector++)
                    {
                        yield return new Damage()
                            {
                                Year = year+1,
                                Region = region+1,
                                Sector = sector,
                                DamageValue = _damages[year,region,sector]
                            };
                    }
                }
            }
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            throw new System.NotImplementedException();
        }
    }

    // GDP for each year and region in US$
    public class Incomes
    {
        private Esmf.JaggedArrayWrapper2<double> m_incomes = new Esmf.JaggedArrayWrapper2<double>(LegacyConstants.NYear, LegacyConstants.NoReg);
        public double this[int Year, int Region] { get { return GetItem(Year, Region); } }

        public void Add(int Year, int Region, double Income)
        {
            m_incomes[Year - 1, Region - 1] = Income;
        }

        public Incomes()
        {
        }

        public double GetItem(int Year, int Region)
        {
            return m_incomes[Year - 1, Region - 1];
        }

    }

    // Population for each year and region in absolute numbers
    public class Populations
    {
        private Esmf.JaggedArrayWrapper2<double> m_populations = new Esmf.JaggedArrayWrapper2<double>(LegacyConstants.NYear, LegacyConstants.NoReg);
        public double this[int Year, int Region] { get { return GetItem(Year, Region); } }

        public void Add(int Year, int Region, double Population)
        {
            m_populations[Year - 1, Region - 1] = Population;
        }

        public Populations()
        {
        }

        public double GetItem(int Year, int Region)
        {
            return m_populations[Year - 1, Region - 1];
        }

    }

}